
    // -@h_densegrid: dense grid setups (coordinates); used for further visualization or registration

    // @h_points : input coords of points
    // @h_sparsestick: input sparse stick, eigen values and eigen vectors
    // @maxGridSize: input maximum possible number of cells in Grid
    // @sigma : input sigma size
    // @cell_size: input cell size of grid
    // @hw_size : input half_window size <- number of cells needs to be extended per point
    // @numPoints : input number of points
    // @min_x(yz): min coord in meter (origin of the grid)
    // @h_logg : output log
    extern "C"
    void CudaVoting::unifiedDenseGridStickVoting(
                const float3 * h_points, 
                const size_t maxGridSize, const int hw_size,
                const float sigma, const float cell_size, const int numPoints, 
                const unsigned int grid_dimx,
                const unsigned int grid_dimy,
                const unsigned int grid_dimz,
                const float min_x,
                const float min_y,
                const float min_z,
                float * stick_saliency_grid,
                float * plate_saliency_grid,
                float * ball_saliency_grid,
                int2 * hh_log)
    {
        // events for timing 
        cudaEvent_t start_event, start_event0, stop_event, stop_event0;
        cutilSafeCall( cudaEventCreate(&start_event) );
        cutilSafeCall( cudaEventCreate(&stop_event) );
        cutilSafeCall( cudaEventCreate(&start_event0) );
        cutilSafeCall( cudaEventCreate(&stop_event0) );
        float elapsed_time0, elapsed_time;
        cudaEventRecord(start_event0, 0);

        // select device and initialize CUDA environment
        cuInit(0);
#ifdef check_memory_size
        // check whether the field size to be created fit in mem
        size_t memsize = sizeof(float)*3*4*numPoints; //(3x3 + 3x1 = 3x4)
        CUdevice cudaDevice;  
        CUresult result = cuDeviceGet(&cudaDevice, 0);  
        if(result!= CUDA_SUCCESS)  
        {  
            cout << "Error fetching cuda device";  
            return ;  
        }  
        CUcontext cudaContext;  
        result = cuCtxCreate(&cudaContext, CU_CTX_SCHED_AUTO, cudaDevice);  
        if(result != CUDA_SUCCESS)  
        {  
            cout << "Error creating cuda context";  
            return ;  
        }
        size_t freeS, totalS;
        cuMemGetInfo(&freeS, &totalS);
        printf("GPU: freeMem = %d , totalMem = %d\n", freeS, totalS);
        if (memsize < freeS)
          printf ("Memory allocation size is ok. Require: %d\n", memsize);
        else{
            printf ("Not enough video memory! Return. Require: d\n", memsize);
            cuCtxDetach( cudaContext );
            return;
        }
#endif

        // start real work

        //
        // create sparse tensor field
        // size of 3x3 matrices per point
        size_t sizeField = numPoints*3*sizeof(float3);
        // copy it to device
        float3 *ds_fieldarray;
        cutilSafeCall( cudaMalloc((void **)&ds_fieldarray, sizeField) );
        cutilSafeCall( cudaMemset(ds_fieldarray, 0, sizeField) );

        //
        // port points as float3
        size_t sizePoints = numPoints*sizeof(float3);
        float3 * d_points;
        cutilSafeCall( cudaMalloc((void **)&d_points, sizePoints) );
        cutilSafeCall( cudaMemcpyAsync(d_points, h_points, sizePoints, cudaMemcpyHostToDevice, 0) );

        size_t sizeLog = numPoints*sizeof(int2);
        int2 * d_log; // [id, as_voter_#votees, as_votee_#voters]
        cutilSafeCall( cudaMalloc((void **)&d_log, sizeLog) );
        cutilSafeCall( cudaMemset(d_log, 0, sizeLog) );


        // ===============================
        // do sparse voting
        cudaEventRecord(stop_event0, 0);
        cudaEventSynchronize(stop_event0);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event0, stop_event0) );
        printf("EVENT timed time before launching kernel :sparse_ball_voting_kernel()\t%.2f \n", elapsed_time );
        printf ("[GPU] deisgned size: %d\n", (numPoints-1)/BLOCK_DIM + 1);
        cudaEventRecord(start_event, 0);
        sparse_ball_voting_kernel <<< (numPoints-1)/BLOCK_DIM + 1, BLOCK_DIM >>> (ds_fieldarray, d_points, sigma, numPoints, d_log);
        cutilCheckMsg("sparse_ball_voting_kernel() execution failed\n");
        cudaEventRecord(stop_event, 0);
        cudaEventSynchronize(stop_event);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event, stop_event) );
        printf("EVENT timed time of kernel 1:sparse_ball_voting_kernel()\t%.2f \n", elapsed_time );
        

        // ===============================
        // do tensor split
        float3 * d_stick_field, * d_plate_field;
        size_t sizeSparseEV = numPoints*sizeof(float3);
        cutilSafeCall( cudaMalloc((void **)&d_stick_field, sizeSparseEV) );
        cutilSafeCall( cudaMalloc((void **)&d_plate_field, sizeSparseEV) );
        float * d_stick, * d_plate, * d_ball;
        size_t sizeSaliency = numPoints*sizeof(float);
        cutilSafeCall( cudaMalloc((void **)&d_stick, sizeSaliency) );
        cutilSafeCall( cudaMalloc((void **)&d_plate, sizeSaliency) );
        cutilSafeCall( cudaMalloc((void **)&d_ball, sizeSaliency) );
        cudaEventRecord(stop_event0, 0);
        cudaEventSynchronize(stop_event0);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event0, stop_event0) );
        printf("EVENT timed time before launching kernel :\t%.2f \n", elapsed_time );
        printf ("[GPU] deisgned size: %d\n", (numPoints-1)/BLOCK_DIM + 1);
        cudaEventRecord(start_event, 0);
        tensor_split_with_eigenvectors_kernel <<< (numPoints-1)/BLOCK_DIM + 1, BLOCK_DIM >>> (ds_fieldarray, d_stick, d_plate, d_ball,
                    d_stick_field, d_plate_field, numPoints);
        cutilCheckMsg("tensor_split_kernel() execution failed\n");
        cudaEventRecord(stop_event, 0);
        cudaEventSynchronize(stop_event);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event, stop_event) );
        printf("EVENT timed time of kernel tensor_split_kernel():\t%.2f \n", elapsed_time );


        // ===============================
        // clean un-needed sparse field
        cutilSafeCall( cudaFree(ds_fieldarray) );

        // ===============================
        // create grid for points
        size_t sizePointGrid = numPoints*sizeof(float3);
        float3 * d_pointgrid; // kept on GPU
        cutilSafeCall( cudaMalloc((void **)&d_pointgrid, sizePointGrid) );
        cutilSafeCall( cudaMemset(d_pointgrid, 0, sizePointGrid) );
        // create grid for the points; get the number of cells needs to be created
        // create d_pointgrid, saving only the nearest point
        // d_pointgrid: [{int_x, int_y, int_z}]
        printf ("[GPU] deisgned size for kernel d_pointgrid: %d\n", (numPoints-1)/BLOCK_DIM_G + 1);
        cudaEventRecord(start_event, 0);
        real_pointgrid_creation <<< (numPoints-1)/BLOCK_DIM_G + 1, BLOCK_DIM_G >>> (d_pointgrid, d_points, 
                    hw_size, cell_size, numPoints, 
                    min_x, min_y, min_z, d_log);
        cutilCheckMsg("real_dense_grid_creation() execution failed\n");
        cudaEventRecord(stop_event, 0);
        cudaEventSynchronize(stop_event);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event, stop_event) );
        printf("EVENT timed time of kernel d_pointgrid:\t%.2f \n", elapsed_time );

        //=========================================
        // do dense voting based on the grids created
        size_t sizeGridField = maxGridSize*3*sizeof(float3); // change to output size
        float3 * d_fieldarray;
        cutilSafeCall( cudaMalloc((void **)&d_fieldarray, sizeGridField) );
        cutilSafeCall( cudaMemset(d_fieldarray, 0, sizeGridField) );

        printf ("[GPU] deisgned size for kernel real_dense_grid_voting(): %ld\n", (maxGridSize-1)/BLOCK_DIM + 1);
        cudaEventRecord(start_event, 0);
        real_dense_stick_grid_voting <<< (maxGridSize-1)/BLOCK_DIM + 1, BLOCK_DIM >>> 
            (d_fieldarray, d_pointgrid, d_stick_field, d_stick,
             hw_size, cell_size, sigma, numPoints, maxGridSize, 
             grid_dimx, grid_dimy, grid_dimz,
             min_x, min_y, min_z, d_log);
        cutilCheckMsg("real_dense_grid_voting() execution failed\n");
        cudaEventRecord(stop_event, 0);

        cudaEventSynchronize(stop_event);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event, stop_event) );
        printf("EVENT timed time of kernel real_dense_grid_voting():\t%.2f \n", elapsed_time );
        

        // === release memory
        cutilSafeCall( cudaFree(d_stick) );
        cutilSafeCall( cudaFree(d_plate) );
        cutilSafeCall( cudaFree(d_ball) );
        cutilSafeCall( cudaFree(d_points) );
        cutilSafeCall( cudaFree(d_log) );



        //=========================================
        // tensor split on d_fieldarray
        // outputing saliency <- get number of cells in the grid first
        float * dd_stick, * dd_plate, * dd_ball; //on-Device Dense (dd)
        int numReal = maxGridSize; // TODO: find the real one
        sizeSaliency = numReal*sizeof(float);
        cutilSafeCall( cudaMalloc((void **)&dd_stick, sizeSaliency) );
        cutilSafeCall( cudaMalloc((void **)&dd_plate, sizeSaliency) );
        cutilSafeCall( cudaMalloc((void **)&dd_ball, sizeSaliency) );
        cudaEventRecord(stop_event0, 0);
        cudaEventSynchronize(stop_event0);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event0, stop_event0) );
        printf("EVENT timed time before launching kernel :\t%.2f \n", elapsed_time );
        printf ("[GPU] deisgned size: %d\n", (numPoints-1)/BLOCK_DIM + 1);
        cudaEventRecord(start_event, 0);
        tensor_split_kernel<<< (numPoints-1)/BLOCK_DIM + 1, BLOCK_DIM >>> (ds_fieldarray, dd_stick, dd_plate, dd_ball,
                    numPoints);
        cutilCheckMsg("tensor_split_kernel() execution failed\n");
        cudaEventRecord(stop_event, 0);
        cudaEventSynchronize(stop_event);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time, start_event, stop_event) );
        printf("EVENT timed time of kernel tensor_split_kernel():\t%.2f \n", elapsed_time );

        cutilSafeCall( cudaMemcpy(stick_saliency_grid, dd_stick, sizeSaliency, cudaMemcpyDeviceToHost) );
        cutilSafeCall( cudaMemcpy(plate_saliency_grid, dd_plate, sizeSaliency, cudaMemcpyDeviceToHost) );
        cutilSafeCall( cudaMemcpy(ball_saliency_grid, dd_ball, sizeSaliency, cudaMemcpyDeviceToHost) );

        //clean up
        cutilSafeCall( cudaFree(dd_stick) );
        cutilSafeCall( cudaFree(dd_plate) );
        cutilSafeCall( cudaFree(dd_ball) );
        cutilSafeCall( cudaFree(d_fieldarray) );

        cudaEventRecord(stop_event0, 0);
        cudaEventSynchronize(stop_event0);
        cutilSafeCall( cudaEventElapsedTime(&elapsed_time0, start_event0, stop_event0) );
        printf("EVENT timed time everything in dense creation:\t%.2f \n\n", elapsed_time0 );
#ifdef check_memory_size
        cuCtxDetach( cudaContext );
#endif
        cutilSafeCall( cudaDeviceReset() );

    }
