/*
 * =====================================================================================
 *
 *       Filename:  vtk_voting_example.cpp
 *
 *    Description:  test cuda tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  04/18/2012 07:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

//#include "CudaVoting.h"
#include "assert.h"
//#include "vtktool.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
#include "aliases.h"

#include "tensor_voting.h"
#include "occupancygrid.h"
#include "vtktool.h"
#include "CudaVoting.h"

#include "boost/tuple/tuple.hpp"
#include <time.h>
#include <cuda.h>
#include <cutil_math.h>
#include <vector_types.h>
#include <iostream>
#include <fstream>

using namespace PointMatcherSupport;
using namespace std;
using namespace topomap;
using namespace cudavoting;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    if (argc < 4)
    {
      cout << "Usage: ./unified_voting_gpu <input.vtk> <output.vtk> sigma gridsize(1.0)" << endl;
      return 0;
    }
    cout << "Input file name: "<< argv[1];
    string filename(argv[1]);
    string outputname(argv[2]);

    DP input;
    input = PM::loadVTK(filename.c_str());
    cout << "Input size: " << input.features.size()/4 << endl;
    // subsampling here. Because the grid will otherwise too slow.
    PM pm;

    int numPoints = input.features.size()/4;

    // use CPU to vote for planes
    float sigma = 0.5;
    if(EOF == sscanf(argv[3], "%f", &sigma))
    {
        cout << "sigma not valid float." << endl;
        return 0;
    }
    cout << "Using sigma size: " << sigma << endl;

    cout << "Sparse ball voting (GPU)..." << endl;
    Eigen::Matrix<Eigen::Matrix3f, Eigen::Dynamic, 1> sparseField;
    // int
    size_t sizeField = numPoints*3*sizeof(int3);
    int3* h_fieldarray = (int3 *)malloc(sizeField);

    size_t sizePoints = numPoints*sizeof(float3);
    float3 *h_points = (float3 *)malloc(sizePoints);
    float max_x = -1e10, max_y = -1e10, max_z = -1e10;
    float min_x = 1e10, min_y = 1e10, min_z = 1e10;
    cout << "Filling host point structure... " << endl;
    for(unsigned int i = 0; i<numPoints; i++) // get min_max of coord at the same time
    {
        h_points[i].x = input.features(0,i); 
        h_points[i].y = input.features(1,i); 
        h_points[i].z = input.features(2,i); 
        if (input.features(0,i) > max_x) max_x = input.features(0,i);
        if (input.features(1,i) > max_y) max_y = input.features(1,i);
        if (input.features(2,i) > max_z) max_z = input.features(2,i);
        if (input.features(0,i) < min_x) min_x = input.features(0,i);
        if (input.features(1,i) < min_y) min_y = input.features(1,i);
        if (input.features(2,i) < min_z) min_z = input.features(2,i);
    }

    // set log
    size_t sizeLog = numPoints*sizeof(int2);
    int2 * h_log = (int2 *)malloc(sizeLog);
    bzero( h_log, sizeLog);

    /*-----------------------------------------------------------------------------
     *  Dense Grid Parameters
     *-----------------------------------------------------------------------------*/
    float cell_size = 1.0;
    if (argc > 4)
      cell_size = atof(argv[4]);
    cout << "Load vtk as grid......cell_size: " << cell_size << endl;

    float hw = 1.0;
    if (argc > 5)
      hw = atof(argv[5]);
    cout << "Using half window " << hw << " size of sigma . Change it with argv[5]." << endl;
    int hw_size = ceil(hw*sigma/cell_size);
    cout << "hw_size is " << hw_size << endl;

    // number of cells in grid
    //size_t maxGridSize = round( 5000.0 / pow(cell_size, 3) );//625000; //0.1*100*50*10/(0.2*0.2*0.2) ;// TODO: modify with point data
    unsigned int grid_dimx = ceil( (max_x - min_x)/ cell_size ) + 2*hw_size +1; // +1 because two ceil function will create 1 additional cell
    unsigned int grid_dimy = ceil( (max_y - min_y)/ cell_size ) + 2*hw_size +1;
    unsigned int grid_dimz = ceil( (max_z - min_z)/ cell_size ) + 2*hw_size +1;
    size_t maxGridSize = grid_dimx * grid_dimy * grid_dimz;
    cout << "@@@@@@@ grid size is " << grid_dimx << " x " << grid_dimy << " x " << grid_dimz << " = " << maxGridSize << "  @@@@@@@@@@" << endl;
    cout << "@@@@@@@ grid ranged from: \n x: [" << min_x << " ," << max_x << "]\n "
        << "y: [" << min_y << " ," << max_y << "]\n "
        << "z: [" << min_z << " ," << max_z << "]\n";


    /*-----------------------------------------------------------------------------
     *  Dense Grid voting
     *-----------------------------------------------------------------------------*/
    // set log
    int2 * hh_log = (int2 *)malloc(sizeLog);
    bzero( hh_log, sizeLog);

    float* stick_saliency_grid = (float *)malloc(maxGridSize*sizeof(float));
    float* plate_saliency_grid = (float *)malloc(maxGridSize*sizeof(float));
    float* ball_saliency_grid = (float *)malloc(maxGridSize*sizeof(float));

    cout << "Send to GPU dense stick voting... numPoints: "<< numPoints << " gridcells: " << maxGridSize << endl;
    cout << "Send to GPU..." << endl;
    // send to GPU
    CudaVoting::unifiedDenseGridPlateVoting(h_points, maxGridSize, hw_size, sigma, cell_size,
                numPoints, grid_dimx, grid_dimy, grid_dimz, 
                min_x, min_y, min_z, 
                stick_saliency_grid, plate_saliency_grid, ball_saliency_grid,
                hh_log); // h_pointgrid is left on GPU only.

    ofstream logff("log_number_of_whowhom_dense_gpu2.txt");
    for(int i =0; i<numPoints; i++)
    {
        logff << i << " " << hh_log[i].x << " " << hh_log[i].y << endl;
    }
    logff.close();

    // post-processing
    cout << "Organize integer conversion... " << endl;
    cout << "Picking non-zero data." << endl;

    vector< boost::tuple<float, float, float> > vecDenseCells; // save centers in meter of all meaningful dense cells

    vecDenseCells.reserve(maxGridSize);

#ifdef FULL_GRID
    // save and grid selection
    for(unsigned int i = 0; i<maxGridSize; i++)
    {
        // save cell definition
        // coord refering origin
        int z = i / (grid_dimx*grid_dimy) - hw_size;
        int xy = i % (grid_dimx*grid_dimy);
        int x = xy % grid_dimx - hw_size;
        int y = xy / grid_dimx - hw_size;

        float xx = x*cell_size + 0.5*cell_size + min_x;
        float yy = y*cell_size + 0.5*cell_size + min_y;
        float zz = z*cell_size + 0.5*cell_size + min_z;
        vecDenseCells.push_back( boost::make_tuple(xx, yy, zz) );
    }
    cout << endl;
    cout << "maxGridSize : " << maxGridSize << endl << "realGridSize : " << vecDenseCells.size() << endl;

    cout	<< "dense tensor split..." << endl;
    PointMatcher<float>::Matrix stick=PM::Matrix::Zero(1, vecDenseCells.size());
    PointMatcher<float>::Matrix plate=PM::Matrix::Zero(1, vecDenseCells.size());
    PointMatcher<float>::Matrix ball =PM::Matrix::Zero(1, vecDenseCells.size());
    for(unsigned int i=0; i<vecDenseCells.size(); i++)
    {
        stick(i) = stick_saliency_grid[i];
        plate(i) = plate_saliency_grid[i]; 
        ball(i) =   ball_saliency_grid[i];
    }
#else // FULL_GRID
    vector<unsigned int> record; // save VALID cells
    record.reserve(maxGridSize);
    // save and grid selection
    for(unsigned int i = 0; i<maxGridSize; i++)
    {
        // save cell definition
        // coord refering origin
        int z = i / (grid_dimx*grid_dimy) - hw_size;
        int xy = i % (grid_dimx*grid_dimy);
        int x = xy % grid_dimx - hw_size;
        int y = xy / grid_dimx - hw_size;

        float xx = x*cell_size + 0.5*cell_size + min_x;
        float yy = y*cell_size + 0.5*cell_size + min_y;
        float zz = z*cell_size + 0.5*cell_size + min_z;

//        if (stick_saliency_grid[i] >1)
           // || plate_saliency_grid[i] > 1e-2
           //|| ball_saliency_grid[i] > 1e-2)
        if (stick_saliency_grid[i] >1e-2 && plate_saliency_grid[i]>1e-2 && ball_saliency_grid[i] >1e-2)
        {
            vecDenseCells.push_back( boost::make_tuple(xx, yy, zz) );
            record.push_back(i);
        }
    }
    cout << endl;
    cout << "maxGridSize : " << maxGridSize << endl << "realGridSize : " << vecDenseCells.size() << endl;

    cout	<< "dense tensor split..." << endl;
    PointMatcher<float>::Matrix stick=PM::Matrix::Zero(1, vecDenseCells.size());
    PointMatcher<float>::Matrix plate=PM::Matrix::Zero(1, vecDenseCells.size());
    PointMatcher<float>::Matrix ball =PM::Matrix::Zero(1, vecDenseCells.size());
    for(unsigned int i=0; i<vecDenseCells.size(); i++)
    {
        stick(i) = stick_saliency_grid[record[i]];
        plate(i) = plate_saliency_grid[record[i]]; 
        ball(i) =   ball_saliency_grid[record[i]];
    }


#endif
    
    // write densegrid to a vector for visualization
    cout	<< "Written to file: " << outputname << endl;
    VTKTool::dumpGridCell(outputname, vecDenseCells, cell_size, stick, plate, ball);
    // clean up
    free(h_points);

//
    return 1;
}	/* ----------  end of function main  ---------- */
