/*
 * =====================================================================================
 *
 *       Filename:  vtk_stick_dense_gpu.cpp
 *
 *    Description:  test cuda dense stick tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  05/21/2012 21:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

#include "assert.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
#include "pointmatcher/Functions.h"
#include "aliases.h"

#include "tensor_voting.h"
#include "CudaVoting.h"

#include <time.h>
#include <cuda.h>
#include <cutil_math.h>
#include <vector_types.h>
#include <iostream>
#include <fstream>
#define USE_GPU_SPARSE_BALL
#define INT_ENDING 1000000

using namespace PointMatcherSupport;
using namespace std;
using namespace topomap;
using namespace cudavoting;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    DP input;
    if (argc > 1)
    {
        cout << "Input file name: "<< argv[1];
        string filename(argv[1]);


        input = PM::loadVTK(filename.c_str());
        cout << "Input size: " << input.features.size()/4 << endl;

        // subsampling here. Because the grid will otherwise too slow.
        PM pm;
        PM::DataPointsFilter* randSample;
        randSample = pm.DataPointsFilterRegistrar.create("RandomSamplingDataPointsFilter",
                    PM::Parameters({
                        {"prob", "0.05"}                    
                        }));
        input = randSample->filter(input);
        cout << "Sampled size: " << input.features.size()/4 << endl;

        PM::saveVTK(input, "sparseFloor.vtk");
    }
    else
    {
        input = PM::loadVTK("sparseFloor.vtk");
    }

    int numPoints = input.features.size()/4;

    // use CPU to vote for planes
    float sigma = 0.5;
    TensorVoting voter(sigma); 
    cout << "Set point sets..." << endl;
    voter.setPoints(input);
    cout << "Sparse ball voting (GPU)..." << endl;
    Eigen::Matrix<Eigen::Matrix3f, Eigen::Dynamic, 1> sparseField;
    Eigen::Matrix<Eigen::Matrix3f, Eigen::Dynamic, 1> denseField;

    // int
    size_t sizeField = numPoints*3*sizeof(int3);
    int3* h_fieldarray = (int3 *)malloc(sizeField);

    //set input:   getArrayFromMatrix();
    size_t sizePoints = numPoints*sizeof(float3);
    float3 *h_points = (float3 *)malloc(sizePoints);
    for(unsigned int i = 0; i<numPoints; i++)
    {
        h_points[i].x = input.features(0,i); 
        h_points[i].y = input.features(1,i); 
        h_points[i].z = input.features(2,i); 
    }

    // set log
    size_t sizeLog = numPoints*sizeof(int2);
    int2 * h_log = (int2 *)malloc(sizeLog);
    bzero( h_log, sizeLog);

    // call CUDA
    cout << "Send to GPU..." << endl;
    CudaVoting::sparseBallVoting(h_fieldarray, h_points, sigma, numPoints, h_log);
    // post-processing
    sparseField.resize(numPoints, 1);
    for(unsigned int i = 0; i<numPoints; i++)
    {
        Eigen::Matrix3f M;
        M << (float)(h_fieldarray[i*3 + 0].x)/INT_ENDING, (float)(h_fieldarray[i*3 + 0].y)/INT_ENDING, (float)(h_fieldarray[i*3 + 0].z)/INT_ENDING, 
             (float)(h_fieldarray[i*3 + 1].x)/INT_ENDING, (float)(h_fieldarray[i*3 + 1].y)/INT_ENDING, (float)(h_fieldarray[i*3 + 1].z)/INT_ENDING, 
             (float)(h_fieldarray[i*3 + 2].x)/INT_ENDING, (float)(h_fieldarray[i*3 + 2].y)/INT_ENDING, (float)(h_fieldarray[i*3 + 2].z)/INT_ENDING;
        sparseField(i) = M;
    }
    voter.setSparseTensor(sparseField);
    printf("Outputing..");

    ofstream out("log_sparse_gpu.txt");
    for(int i =0; i<numPoints; i++)
        out << sparseField(i) << endl;
    out.close();


    ofstream out1("log_point_gpu.txt");
    out1 << input.features << endl;
    out1.close();

    ofstream logf("log_number_of_whowhom_gpu.txt");
    for(int i =0; i<numPoints; i++)
    {
        logf << i << " " << h_log[i].x << " " << h_log[i].y << endl;
    }
    logf.close();


    // clean up
    free(h_fieldarray);
    free(h_log);

    cout	<< "sparse tensor split..." << endl;
    voter.sparse_tensor_split();

    // save sparse tensor
    PM::DataPoints::Descriptors stick=PM::DataPoints::Descriptors::Zero(1, numPoints);
    PM::DataPoints::Descriptors plate=PM::DataPoints::Descriptors::Zero(1, numPoints);
    PM::DataPoints::Descriptors ball =PM::DataPoints::Descriptors::Zero(1, numPoints);
    for(unsigned int i=0; i<numPoints; i++)
    {
        stick(i) = voter.sparseStick(0,i);
        plate(i) = voter.sparsePlate(0,i*2);
        ball(i) =  voter.sparseBall(i);
    }
    // TODO: enable it when need the sparse
    //input.addDescriptor("stick", stick);
    //input.addDescriptor("plate", plate);
    //input.addDescriptor("ball", ball);
    //PM::saveVTK(input, "outputSparseGPU.vtk");

    //
    // Start dense stick voting
    //

    // setZero of h_field
    int3 * h_fieldarraydense = (int3 *)malloc(sizeField);
    bzero( h_fieldarraydense, sizeField );

    // copy to get h_sparsestick as input to GPU
    size_t sizeSparseStick = numPoints*4*sizeof(float4);
    float4 * h_sparsestick = (float4 *)malloc(sizeSparseStick);
    for(unsigned int i = 0; i<numPoints; i++)
    {
        h_sparsestick[i].x = voter.sparseStick(0,i); 
        h_sparsestick[i].y = voter.sparseStick(1,i); 
        h_sparsestick[i].z = voter.sparseStick(2,i); 
        h_sparsestick[i].w = voter.sparseStick(3,i); 
    }
    // send to GPU
    cout << "Send to GPU dense stick voting..." << endl;
    CudaVoting::denseStickVoting(h_fieldarraydense, h_points, h_sparsestick, sigma, numPoints);
    // post-processing
    denseField.resize(numPoints, 1);
    for(unsigned int i = 0; i<numPoints; i++)
    {
        Eigen::Matrix3f M;
        M << (float)(h_fieldarraydense[i*3 + 0].x)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 0].y)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 0].z)/INT_ENDING, 
             (float)(h_fieldarraydense[i*3 + 1].x)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 1].y)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 1].z)/INT_ENDING, 
             (float)(h_fieldarraydense[i*3 + 2].x)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 2].y)/INT_ENDING, (float)(h_fieldarraydense[i*3 + 2].z)/INT_ENDING;
        denseField(i) = M;
    }
    // TODO: setDenseTensor()
    //voter.setDenseTensor(denseField);
    voter.setSparseTensor(denseField);
    printf("Outputing..");

    ofstream outdense("log_dense_gpu.txt");
    for(int i =0; i<numPoints; i++)
        outdense << denseField(i) << endl;
    outdense.close();

    // clean up
    free(h_fieldarraydense);
    free(h_sparsestick);
    free(h_points);

    cout	<< "dense tensor split..." << endl;
    voter.sparse_tensor_split();

    cout	<< "Split done." << endl;
    // save dense tensor
    stick=PM::DataPoints::Descriptors::Zero(1, numPoints);
    plate=PM::DataPoints::Descriptors::Zero(1, numPoints);
    ball =PM::DataPoints::Descriptors::Zero(1, numPoints);
    //TODO: after impl of setDenseTensor(), use splitted denseStick
    cout	<< "Save Split tensors." << endl;
    for(unsigned int i=0; i<numPoints; i++)
    {
        stick(i) = voter.sparseStick(0,i);
        plate(i) = voter.sparsePlate(0,i*2);
        ball(i) =  voter.sparseBall(i);
    }
    input.addDescriptor("stick", stick);
    input.addDescriptor("plate", plate);
    input.addDescriptor("ball", ball);

    cout	<< "Write to file." << endl;
    PM::saveVTK(input, "outputDenseGPU.vtk");
    return 1;
}	/* ----------  end of function main  ---------- */
