/*
 * =====================================================================================
 *
 *       Filename:  vtk_voting_example.cpp
 *
 *    Description:  test cuda tensor voting library with vtk input
 *
 *        Version:  1.0
 *        Created:  04/18/2012 07:55 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ming Liu (), ming.liu@mavt.ethz.ch
 *        Company:  ETHZ
 *
 * =====================================================================================
 */

#include <iostream>

//#include "CudaVoting.h"
#include "assert.h"
//#include "vtktool.h"
// libpointmatcher
#include "pointmatcher/PointMatcher.h"
#include "pointmatcher/Functions.h"
#include "aliases.h"

#include "tensor_voting.h"

#include <time.h>

#define USE_GPU_SPARSE_BALL

using namespace PointMatcherSupport;
using namespace std;
using namespace topomap;
//using namespace cudavoting;

//! A dense matrix over ScalarType
typedef Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> Matrix;
//! A dense integer matrix
typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic> IntMatrix;

/* 
 * ===  FUNCTION  ============================================================
 *         Name:  Main  
 *  Description:  Entry of program
 * ===========================================================================
 */
int main ( int argc, char *argv[] )
{
    cout << "Input file name: "<< argv[1];
    string filename(argv[1]);


    DP input;
    input = PM::loadVTK(filename.c_str());
    cout << "Input size: " << input.features.size()/4 << endl;

    // subsampling here. Because the grid will otherwise too slow.
    PM pm;
    PM::DataPointsFilter* randSample;
    randSample = pm.DataPointsFilterRegistrar.create("RandomSamplingDataPointsFilter",
                PM::Parameters({
                    {"prob", "0.05"}                    
                    }));
    input = randSample->filter(input);
    cout << "Sampled size: " << input.features.size()/4 << endl;

    int numPoints = input.features.size()/4;

    // use CPU to vote for planes
    float sigma = 0.5;
    TensorVoting voter(sigma); 
    cout << "Set point sets..." << endl;
    voter.setPoints(input);
    cout << "Sparse ball voting (CPU)..." << endl;
    voter.sparse_ball_vote();
    cout	<< "sparse tensor split..." << endl;
    voter.sparse_tensor_split();

    Eigen::Matrix<Eigen::Matrix3f, Eigen::Dynamic, 1> field= voter.stickDenseVote();
    vector<TensorEigenValuesType> tensorVector;
    tensorVector.reserve(field.rows());
    /*-----------------------------------------------------------------------------
     *  Organize Tensor voting results
     *-----------------------------------------------------------------------------*/
    for(unsigned int i=0; i<field.rows(); i++)
    {
        printf("Extract eigen values: %d / %d\r", i, field.rows());
        Eigen::SelfAdjointEigenSolver<Eigen::Matrix3f> es(field(i), false); // only calculate eigen values
        Eigen::Vector3f abs_evalues;
        Eigen::Vector3i eig_indices(0,0,0);

        abs_evalues = es.eigenvalues().array().abs();
        abs_evalues.maxCoeff(&eig_indices(0));
        abs_evalues.minCoeff(&eig_indices(2));
        for (unsigned int x=0; x<3; ++x)
        {
            if (x != eig_indices(0) && x!= eig_indices(2)) { eig_indices(1) = x; }
        }	
        TensorEigenValuesType tensorEigenvalues(
                    es.eigenvalues()[eig_indices(0)] - es.eigenvalues()[eig_indices(1)], // stick <- voter for planes
                    es.eigenvalues()[eig_indices(1)] - es.eigenvalues()[eig_indices(2)], // plate <- voter for curves
                    es.eigenvalues()[eig_indices(2)] // ball <- voter for single points
                    ); 

        tensorVector.push_back(tensorEigenvalues);
        //printf("where: %d \t Stick: %.3f \t Plate: %.3f \t Ball: %.3f\n", 
        //            i, tensorEigenvalues.get<0>(), tensorEigenvalues.get<1>(), tensorEigenvalues.get<2>());
    }
    PM::DataPoints::Descriptors stick=PM::DataPoints::Descriptors::Zero(1, field.rows());
    PM::DataPoints::Descriptors plate=PM::DataPoints::Descriptors::Zero(1, field.rows());
    PM::DataPoints::Descriptors ball =PM::DataPoints::Descriptors::Zero(1, field.rows());
    for(unsigned int i=0; i<tensorVector.size(); i++)
    {
        stick(i) = tensorVector[i].get<0>();
        plate(i) = tensorVector[i].get<1>();
        ball(i) = tensorVector[i].get<2>();
    }
    input.addDescriptor("stick", stick);
    input.addDescriptor("plate", plate);
    input.addDescriptor("ball", ball);

    PM::saveVTK(input, "output.vtk");

    return 1;
}	/* ----------  end of function main  ---------- */
